@@ -94,10 +94,10 @@ pub fn parse_obj<T: BufRead>(input: T) -> ObjResult<RawObj> {
     let mut polygons = Vec::new();
 
     let counter = Counter::new(&points, &lines, &polygons);
-    let mut group_builder = counter.hash_map("default".to_string());
-    let mut mesh_builder = counter.hash_map(String::new());
-    let mut smoothing_builder = counter.vec_map();
-    let mut merging_builder = counter.vec_map();
+    let mut group_builder = GroupBuilder::with_default(&counter, String::from("default"));
+    let mut mesh_builder = GroupBuilder::with_default(&counter, String::new());
+    let mut smoothing_builder = GroupBuilder::new(&counter);
+    let mut merging_builder = GroupBuilder::new(&counter);
 
     lex(input, |stmt, args: &[&str]| {
         match stmt {
@@ -213,7 +213,7 @@ pub fn parse_obj<T: BufRead>(input: T) -> ObjResult<RawObj> {
 
             // Grouping
             "g" => match args {
-                [name] => group_builder.start((*name).to_string()),
+                [name] => group_builder.try_start((*name).to_string())?,
                 _ => make_error!(
                     WrongNumberOfArguments,
                     "Expected group name parameter, but nothing has been supplied"
@@ -221,12 +221,12 @@ pub fn parse_obj<T: BufRead>(input: T) -> ObjResult<RawObj> {
             },
             "s" => match args {
                 ["off"] | ["0"] => smoothing_builder.end(),
-                [param] => smoothing_builder.start(param.parse()?),
+                [param] => smoothing_builder.try_start(param.parse()?)?,
                 _ => make_error!(WrongNumberOfArguments, "Expected only 1 argument"),
             },
             "mg" => match args {
                 ["off"] | ["0"] => merging_builder.end(),
-                [param] => merging_builder.start(param.parse()?),
+                [param] => merging_builder.try_start(param.parse()?)?,
                 _ => make_error!(WrongNumberOfArguments, "Expected only 1 argument"),
             },
             "o" => {
@@ -243,7 +243,7 @@ pub fn parse_obj<T: BufRead>(input: T) -> ObjResult<RawObj> {
             "d_interp" => unimplemented!(),
             "lod" => unimplemented!(),
             "usemtl" => match args {
-                [material] => mesh_builder.start((*material).to_string()),
+                [material] => mesh_builder.try_start((*material).to_string())?,
                 _ => make_error!(WrongNumberOfArguments, "Expected only 1 argument"),
             },
             "mtllib" => {
@@ -325,27 +325,6 @@ impl Counter {
             )
         }
     }
-
-    /// Creates a `HashMap<String, Group>` builder which references `self` as counter.
-    fn hash_map(&self, input: String) -> GroupBuilder<'_, HashMap<String, Group>, String> {
-        let mut result = HashMap::with_capacity(1);
-        result.insert(input.clone(), Group::new((0, 0, 0)));
-
-        GroupBuilder {
-            counter: self,
-            current: Some(input),
-            result,
-        }
-    }
-
-    /// Creates a `VecMap<Group>` builder which references `self` as counter.
-    fn vec_map(&self) -> GroupBuilder<'_, VecMap<Group>, usize> {
-        GroupBuilder {
-            counter: self,
-            current: None,
-            result: VecMap::new(),
-        }
-    }
 }
 
 /// Helper for creating `groups`, `meshes`, `smoothing_groups` and `merging_groups` member of
@@ -362,27 +341,47 @@ where
     T: Map<K, Group>,
     K: Clone + Key,
 {
+    fn new(counter: &'a Counter) -> Self {
+        GroupBuilder {
+            counter,
+            current: None,
+            result: T::new(),
+        }
+    }
+
+    fn with_default(counter: &'a Counter, default: K) -> Self {
+        let mut result = T::with_capacity(1);
+        result
+            .try_insert(default.clone(), Group::new((0, 0, 0)))
+            .unwrap();
+
+        GroupBuilder {
+            counter,
+            current: Some(default),
+            result,
+        }
+    }
+
     /// Starts a group whose name is `input`.
-    fn start(&mut self, input: K) {
+    fn try_start(&mut self, input: K) -> ObjResult<()> {
         let count = self.counter.get();
         if let Some(ref current) = self.current {
             if *current == input {
-                return;
+                return Ok(());
             }
             if self.result.get_mut(current).unwrap().end(count) {
                 let res = self.result.remove(&current);
                 assert!(res.is_some());
             }
         }
-        (|| {
-            if let Some(ref mut group) = self.result.get_mut(&input) {
-                group.start(count);
-                return;
-            }
-            let res = self.result.insert(input.clone(), Group::new(count));
+        if let Some(ref mut group) = self.result.get_mut(&input) {
+            group.start(count);
+        } else {
+            let res = self.result.try_insert(input.clone(), Group::new(count))?;
             assert!(res.is_none());
-        })();
+        }
         self.current = Some(input);
+        Ok(())
     }
 
     /// Ends a current group.
@@ -455,8 +454,10 @@ impl Group {
 
 /// Custom trait to interface `HashMap` and `VecMap`.
 trait Map<K: Key, V> {
+    fn new() -> Self;
+    fn with_capacity(capacity: usize) -> Self;
     /// Interface of `insert` function.
-    fn insert(&mut self, _: K, _: V) -> Option<V>;
+    fn try_insert(&mut self, _: K, _: V) -> ObjResult<Option<V>>;
     /// Interface of `get_mut` function.
     fn get_mut(&mut self, k: &K) -> Option<&mut V>;
     /// Interface of `remove` function.
@@ -464,8 +465,14 @@ trait Map<K: Key, V> {
 }
 
 impl<V> Map<String, V> for HashMap<String, V> {
-    fn insert(&mut self, k: String, v: V) -> Option<V> {
-        self.insert(k, v)
+    fn new() -> Self {
+        HashMap::new()
+    }
+    fn with_capacity(capacity: usize) -> Self {
+        HashMap::with_capacity(capacity)
+    }
+    fn try_insert(&mut self, k: String, v: V) -> ObjResult<Option<V>> {
+        Ok(self.insert(k, v))
     }
     fn get_mut(&mut self, k: &String) -> Option<&mut V> {
         self.get_mut(k)
@@ -476,8 +483,24 @@ impl<V> Map<String, V> for HashMap<String, V> {
 }
 
 impl<V> Map<usize, V> for VecMap<V> {
-    fn insert(&mut self, k: usize, v: V) -> Option<V> {
-        self.insert(k, v)
+    fn new() -> Self {
+        VecMap::new()
+    }
+    fn with_capacity(capacity: usize) -> Self {
+        VecMap::with_capacity(capacity)
+    }
+    fn try_insert(&mut self, k: usize, v: V) -> ObjResult<Option<V>> {
+        use crate::error::{LoadError, LoadErrorKind, ObjError};
+
+        // Biggest group number will be 67108864, which is large enough.
+        // It will consume 64MiB of memory in the worst case.
+        if k > 64 * (1 << 20) {
+            return Err(ObjError::Load(LoadError::new(
+                LoadErrorKind::TooBigGroupNumber,
+                "Currently biggest group number allowed is 67108864.",
+            )));
+        }
+        Ok(self.insert(k, v))
     }
     fn get_mut(&mut self, k: &usize) -> Option<&mut V> {
         self.get_mut(*k)
